
 /**
  * @package subtle
  *
  * @file subtle ruby extension
  * @copyright (c) 2005-2009 Christoph Kappel <unexist@dorfelite.net>
  * @version $Id: src/subtlext/subtlext.c,v 1383 2009/09/03 07:47:22 unexist $
  *
  * This program can be distributed under the terms of the GNU GPL.
  * See the file COPYING.
  **/

#include <ruby.h>
#include "shared.h"

#ifdef HAVE_X11_EXTENSIONS_XINERAMA_H
#include <X11/extensions/Xinerama.h>
#endif /* HAVE_X11_EXTENSIONS_XINERAMA_H */

Display *display = NULL;
static int refcount = 0;
VALUE mod = Qnil;

#ifdef DEBUG
int debug = 0;
#endif /* DEBUG */

/* Prototypes {{{ */
static VALUE SubtlextSubtleClientCurrent(VALUE self);
static VALUE SubtlextClientUpdate(VALUE self);
/* }}} */

/* Flags {{{ */
#define SUB_TYPE_CLIENT  0           ///< Client
#define SUB_TYPE_VIEW    1           ///< View
#define SUB_TYPE_TAG     2           ///< Tag
#define SUB_TYPE_SCREEN  3           ///< Screen

#define SUB_ACTION_TAG   0           ///< Tag
#define SUB_ACTION_UNTAG 1           ///< Untag
/* }}} */

/* SubtlextFind {{{ */
static VALUE
SubtlextFind(int type,
  VALUE value,
  int create)
{
  int id = -1;
  Window win = -1;
  VALUE klass = Qnil, obj = Qnil, name = Qnil;
  SubMessageData data = { { 0, 0, 0, 0, 0 } };

  assert(SUB_TYPE_TAG == type || SUB_TYPE_VIEW == type || SUB_TYPE_SCREEN == type);

  switch(type)
    {
      case SUB_TYPE_TAG:    name = rb_intern("Tag");    break;
      case SUB_TYPE_VIEW:   name = rb_intern("View");   break;
      case SUB_TYPE_SCREEN: name = rb_intern("Screen"); break;
    }

  klass = rb_const_get(mod, name); 

  /* Check object type */
  switch(rb_type(value))
    {
      case T_FIXNUM:
        if(SUB_TYPE_SCREEN == type)
          {
            int n = 0;
            
#ifdef HAVE_X11_EXTENSIONS_XINERAMA_H
            int xinerama_event = 0, xinerama_error = 0;
#endif /* HAVE_X11_EXTENSIONS_XINERAMA_H */

            id  = FIX2INT(value);
            obj = rb_funcall(klass, rb_intern("new"), 1, value);

#ifdef HAVE_X11_EXTENSIONS_XINERAMA_H
            /* Xinerama */
            if(XineramaQueryExtension(display, &xinerama_event, &xinerama_error) &&
              XineramaIsActive(display))
              {
                XineramaScreenInfo *screens = NULL;

                /* Query screens */
                if((screens = XineramaQueryScreens(display, &n)))
                  {
                    if(0 <= id && n > id)
                      {
                        rb_iv_set(obj, "@x",      INT2FIX(screens[id].x_org));
                        rb_iv_set(obj, "@y",      INT2FIX(screens[id].y_org));
                        rb_iv_set(obj, "@width",  INT2FIX(screens[id].width));
                        rb_iv_set(obj, "@height", INT2FIX(screens[id].height));
                      }
                    else 
                      {
                        rb_raise(rb_eStandardError, "Failed finding screen");

                        return Qnil;
                      }

                    XFree(screens);
                  }
              } 
#endif /* HAVE_X11_EXTENSIONS_XINERAMA_H */

            /* Return default screen */
            if(0 == n)
              {
                rb_iv_set(obj, "@x",      INT2FIX(0));
                rb_iv_set(obj, "@y",      INT2FIX(0));
                rb_iv_set(obj, "@width",  INT2FIX(DisplayWidth(display, DefaultScreen(display))));
                rb_iv_set(obj, "@height", INT2FIX(DisplayHeight(display, DefaultScreen(display))));
              }

            return obj;
          }
        break;
      case T_STRING:
        if(SUB_TYPE_TAG == type) id = subSharedTagFind(RSTRING_PTR(value));
        else id = subSharedViewFind(RSTRING_PTR(value), &win);

        /* Create and find again */
        if(-1 == id && True == create)
          {
            snprintf(data.b, sizeof(data.b), "%s", RSTRING_PTR(value));
            subSharedMessage(DefaultRootWindow(display), 
              SUB_TYPE_TAG == type ? "SUBTLE_TAG_NEW" : "SUBTLE_VIEW_NEW", data, True);

            if(SUB_TYPE_TAG == type) id = subSharedTagFind(RSTRING_PTR(value));
            else id = subSharedViewFind(RSTRING_PTR(value), &win);
          }

        if(-1 != id)
          {
            /* Create new instance */
            obj = rb_funcall(klass, rb_intern("new"), 1, value);

            rb_iv_set(obj, "@id",   INT2FIX(id));
            rb_iv_set(obj, "@name", value);
            if(SUB_TYPE_VIEW == type) rb_iv_set(obj, "@win", LONG2NUM(win));

            return obj;
          }

        rb_raise(rb_eStandardError, "Failed finding %s", SUB_TYPE_TAG == type ? "tag" : "view");
        return Qnil;
      case T_OBJECT:
      case T_CLASS:
        if(rb_obj_is_instance_of(value, klass)) ///< Check object instance
          {
            if(Qnil == rb_iv_get(value, "@id") && True == create)
              {
                name = rb_iv_get(value, "@name");

                /* Inform subtle */
                snprintf(data.b, sizeof(data.b), "%s", RSTRING_PTR(name));
                subSharedMessage(DefaultRootWindow(display), 
                  SUB_TYPE_TAG == type ? "SUBTLE_TAG_NEW" : "SUBTLE_VIEW_NEW", data, True);

                /* Check if object exists */
                if(SUB_TYPE_TAG == type) id = subSharedTagFind(RSTRING_PTR(name));
                else id = subSharedViewFind(RSTRING_PTR(name), &win);
                if(-1 == id)
                  {
                    rb_raise(rb_eStandardError, "Failed updating %s", 
                      SUB_TYPE_TAG == type ? "tag" : "view");
                    return Qnil;
                  }                

                /* Update object */
                rb_iv_set(value, "@id", INT2FIX(id)); 
                if(SUB_TYPE_VIEW == type) rb_iv_set(value, "@win", LONG2NUM(win));
              }

            return value;
          }
    }

  rb_raise(rb_eArgError, "Unknown value type");
  return Qnil;
} /* }}} */

/* SubtlextToggle {{{ */
static VALUE
SubtlextToggle(VALUE self,
  char *type,
  int flag)
{
  Window win = 0;

  if((win = NUM2LONG(rb_iv_get(self, "@win"))))
    {
      int flags = 0;
      SubMessageData data = { { 0, 0, 0, 0, 0 } };

      flags     = FIX2INT(rb_iv_get(self, "@flags"));
      data.l[1] = XInternAtom(display, type, False);

      /* Toggle flag */
      if(flags & flag) flags &= ~flag;
      else flags |= flag;

      rb_iv_set(self, "@flags", INT2FIX(flags));

      subSharedMessage(win, "_NET_WM_STATE", data, True);
    }
  else rb_raise(rb_eStandardError, "Failed toggling client");

  return Qnil;
} /* }}} */

/* SubtlextRestack {{{ */
static VALUE
SubtlextRestack(VALUE self,
  int detail)
{
  VALUE win = rb_iv_get(self, "@win");
  SubMessageData data = { { 0, 0, 0, 0, 0 } };

  if(RTEST(win))
    {
      data.l[0] = 2; ///< Claim to be a pager
      data.l[1] = NUM2LONG(win);
      data.l[2] = detail; 

      subSharedMessage(DefaultRootWindow(display), "_NET_RESTACK_WINDOW", data, True);
    }
  else rb_raise(rb_eStandardError, "Failed restacking client");  

  return Qnil;
} /* }}} */

/* SubtlextClientFromWin {{{ */
static VALUE
SubtlextClientFromWin(Window win)
{
  VALUE klass = Qnil, client = Qnil;

  /* Create new instance */
  klass  = rb_const_get(mod, rb_intern("Client"));
  client = rb_funcall(klass, rb_intern("new"), 1, LONG2NUM(win));

  SubtlextClientUpdate(client);

  return client;
} /* }}} */

/* SubtlextMatch {{{ */
static VALUE
SubtlextMatch(VALUE self,
  int type)
{
  int i, id = 0, size = 0, match = 0, score = 0, *gravity1 = NULL;
  Window *clients = NULL, *views = NULL, found = None;
  VALUE win = Qnil, client = Qnil;
  unsigned long *cv = NULL, *flags1 = NULL;

  win     = rb_iv_get(self, "@win");
  clients = subSharedClientList(&size);
  views   = (Window *)subSharedPropertyGet(DefaultRootWindow(display), XA_WINDOW,
    "_NET_VIRTUAL_ROOTS", NULL);
  cv      = (unsigned long *)subSharedPropertyGet(DefaultRootWindow(display),
    XA_CARDINAL, "_NET_CURRENT_DESKTOP", NULL);

  if(clients && cv)
    {
      flags1   = (unsigned long *)subSharedPropertyGet(views[*cv], XA_CARDINAL,
        "SUBTLE_WINDOW_TAGS", NULL);
      gravity1 = (int *)subSharedPropertyGet(NUM2LONG(win), XA_CARDINAL,
        "SUBTLE_WINDOW_GRAVITY", NULL);

      /* Iterate once to find a client score-based */
      for(i = 0; 100 != match && i < size; i++)
        {
          unsigned long *flags2 = (unsigned long *)subSharedPropertyGet(clients[i], XA_CARDINAL,
            "SUBTLE_WINDOW_TAGS", NULL);

          if(win != clients[i] && *flags1 & *flags2) ///< Check if there are common tags
            {
              int *gravity2 = (int *)subSharedPropertyGet(clients[i], XA_CARDINAL,
                "SUBTLE_WINDOW_GRAVITY", NULL);

              if(match < (score = subSharedMatch(type, *gravity1, *gravity2)))
                {
                  match = score;
                  found = clients[i];
                  id    = i;
                }

              free(gravity2);
            }

          free(flags2);
        }

      if(found) client = SubtlextClientFromWin(found);

      free(flags1);
      free(gravity1);
      free(clients);
      free(cv);
    }

  return client;
} /* }}} */

/* SubtlextTag {{{ */
static VALUE
SubtlextTag(VALUE self,
  VALUE value,
  int action,
  int type)
{
  VALUE tag = Qnil;

  /* Find tag */
  if(Qnil != (tag = SubtlextFind(SUB_TYPE_TAG, value, True)))
    {
      VALUE oid = Qnil, tid = Qnil;

      oid = rb_iv_get(self, "@id");
      tid = rb_iv_get(tag, "@id");

      if(Qnil != oid && Qnil != tid)
        {
          SubMessageData data = { { 0, 0, 0, 0, 0 } };

          data.l[0] = FIX2LONG(oid);
          data.l[1] = FIX2LONG(tid);
          data.l[2] = type;

          subSharedMessage(DefaultRootWindow(display),
            SUB_ACTION_TAG == action ? "SUBTLE_WINDOW_TAG" : "SUBTLE_WINDOW_UNTAG",
            data, True);

          return Qtrue;
        }
    }

  rb_raise(rb_eArgError, "Unknown value type");
  return Qfalse;
} /* }}} */

/* SubtlextScreens {{{ */
static VALUE
SubtlextScreens(void)
{
  int n = 0;
  VALUE method = Qnil, klass = Qnil, array = Qnil; 

#ifdef HAVE_X11_EXTENSIONS_XINERAMA_H
  int xinerama_event = 0, xinerama_error = 0;
#endif /* HAVE_X11_EXTENSIONS_XINERAMA_H */

  method = rb_intern("new");
  klass  = rb_const_get(mod, rb_intern("Screen"));
  array  = rb_ary_new();

#ifdef HAVE_X11_EXTENSIONS_XINERAMA_H
  /* Xinerama */
  if(XineramaQueryExtension(display, &xinerama_event, &xinerama_error) &&
    XineramaIsActive(display))
    {
      int i;
      XineramaScreenInfo *screens = NULL;

      /* Query screens */
      if((screens = XineramaQueryScreens(display, &n)))
        {
          for(i = 0; i < n; i++)
            {
              VALUE s = rb_funcall(klass, method, 1, INT2FIX(i));

              rb_iv_set(s, "@x",      INT2FIX(screens[i].x_org));
              rb_iv_set(s, "@y",      INT2FIX(screens[i].y_org));
              rb_iv_set(s, "@width",  INT2FIX(screens[i].width));
              rb_iv_set(s, "@height", INT2FIX(screens[i].height));

              rb_ary_push(array, s);
            }

          XFree(screens);
        }
    } 
#endif /* HAVE_X11_EXTENSIONS_XINERAMA_H */

  /* Get default screen */
  if(0 == n)
    {
      VALUE s = rb_funcall(klass, method, 1, INT2FIX(0));

      rb_iv_set(s, "@x",      INT2FIX(0));
      rb_iv_set(s, "@y",      INT2FIX(0));
      rb_iv_set(s, "@width",  INT2FIX(DisplayWidth(display, DefaultScreen(display))));
      rb_iv_set(s, "@height", INT2FIX(DisplayHeight(display, DefaultScreen(display))));

      rb_ary_push(array, s);
    }

  return array;
} /* }}} */

/* SubtlextClientInit {{{ */
static VALUE
SubtlextClientInit(VALUE self,
  VALUE win)
{
  rb_iv_set(self, "@id",      Qnil);
  rb_iv_set(self, "@win",     win);
  rb_iv_set(self, "@name",    Qnil);
  rb_iv_set(self, "@klass",   Qnil);
  rb_iv_set(self, "@x",       Qnil);
  rb_iv_set(self, "@y",       Qnil); 
  rb_iv_set(self, "@width",   Qnil); 
  rb_iv_set(self, "@height",  Qnil); 
  rb_iv_set(self, "@gravity", Qnil);
  rb_iv_set(self, "@screen",  Qnil);
  rb_iv_set(self, "@flags",   Qnil);

  return self;
} /* }}} */

/* SubtlextClientViewList {{{ */
static VALUE
SubtlextClientViewList(VALUE self)
{
  int i, size = 0;
  char **names = NULL;
  Window *views = NULL;
  VALUE win = Qnil, array = Qnil, method = Qnil, klass = Qnil;
  unsigned long *flags1 = NULL;
  
  win     = rb_iv_get(self, "@win");
  method  = rb_intern("new");
  klass   = rb_const_get(mod, rb_intern("View"));
  array   = rb_ary_new2(size);
  names   = subSharedPropertyStrings(DefaultRootWindow(display), 
    "_NET_DESKTOP_NAMES", &size);
  views   = (Window *)subSharedPropertyGet(DefaultRootWindow(display), 
    XA_WINDOW, "_NET_VIRTUAL_ROOTS", NULL);
  flags1  = (unsigned long *)subSharedPropertyGet(NUM2LONG(win), XA_CARDINAL, 
    "SUBTLE_WINDOW_TAGS", NULL);

  if(names && views)
    {
      for(i = 0; i < size; i++)
        {
          unsigned long *flags2 = (unsigned long *)subSharedPropertyGet(views[i], 
            XA_CARDINAL, "SUBTLE_WINDOW_TAGS", NULL);

          if(*flags1 & *flags2) ///< Check if there are common tags
            {
              VALUE v = rb_funcall(klass, method, 1, rb_str_new2(names[i]));

              rb_iv_set(v, "@id",  INT2FIX(i));
              rb_iv_set(v, "@win", LONG2NUM(views[i]));
              rb_ary_push(array, v);
            }

          free(flags2);
        }

      XFreeStringList(names);
      free(views);
    }

  return array;
} /* }}} */

/* SubtlextClientTagList {{{ */
static VALUE
SubtlextClientTagList(VALUE self)
{
  Window win = 0;
  int i, size = 0;
  char **tags = NULL;
  unsigned long *flags = NULL;
  VALUE array = Qnil, method = Qnil, klass = Qnil;

  if((win = NUM2LONG(rb_iv_get(self, "@win"))))
    {
      method = rb_intern("new");
      klass  = rb_const_get(mod, rb_intern("Tag"));
      flags  = (unsigned long *)subSharedPropertyGet(win, XA_CARDINAL, 
        "SUBTLE_WINDOW_TAGS", NULL);
      tags   = subSharedPropertyStrings(DefaultRootWindow(display), "SUBTLE_TAG_LIST", &size);
      array  = rb_ary_new2(size);

      /* Build tag array */
      for(i = 0; i < size; i++)
        {
          if((int)*flags & (1L << (i + 1)))
            {
              VALUE t = rb_funcall(klass, method, 1, rb_str_new2(tags[i]));
              rb_iv_set(t, "@id", INT2FIX(i));
              rb_ary_push(array, t);
            }
        }

      XFreeStringList(tags);
      free(flags);
    }

  return array;
} /* }}} */

/* SubtlextClientTagAdd {{{ */
static VALUE
SubtlextClientTagAdd(VALUE self,
  VALUE value)
{
  return SubtlextTag(self, value, SUB_ACTION_TAG, SUB_TYPE_CLIENT);
} /* }}} */

/* SubtlextClientTagDel {{{ */
static VALUE
SubtlextClientTagDel(VALUE self,
  VALUE value)
{
  return SubtlextTag(self, value, SUB_ACTION_UNTAG, SUB_TYPE_CLIENT);
} /* }}} */

/* SubtlextClientStateFull {{{ */
VALUE
SubtlextClientStateFull(VALUE self)
{
  return FIX2INT(rb_iv_get(self, "@flags")) & SUB_EWMH_FULL ? Qtrue : Qfalse;
} /* }}} */

/* SubtlextClientStateFloat {{{ */
VALUE
SubtlextClientStateFloat(VALUE self)
{
  return FIX2INT(rb_iv_get(self, "@flags")) & SUB_EWMH_FLOAT ? Qtrue : Qfalse;
} /* }}} */

/* SubtlextClientStateStick {{{ */
VALUE
SubtlextClientStateStick(VALUE self)
{
  return FIX2INT(rb_iv_get(self, "@flags")) & SUB_EWMH_STICK ? Qtrue : Qfalse;
} /* }}} */

/* SubtlextClientToggleFull {{{ */
static VALUE
SubtlextClientToggleFull(VALUE self)
{
  return SubtlextToggle(self, "_NET_WM_STATE_FULLSCREEN", SUB_EWMH_FULL);
} /* }}} */

/* SubtlextClientToggleFloat {{{ */
static VALUE
SubtlextClientToggleFloat(VALUE self)
{
  return SubtlextToggle(self, "_NET_WM_STATE_ABOVE", SUB_EWMH_FLOAT);
} /* }}} */

/* SubtlextClientToggleStick {{{ */
static VALUE
SubtlextClientToggleStick(VALUE self)
{
  return SubtlextToggle(self, "_NET_WM_STATE_STICKY", SUB_EWMH_STICK);
} /* }}} */

/* SubtlextClientRestackRaise {{{ */
static VALUE
SubtlextClientRestackRaise(VALUE self)
{
  return SubtlextRestack(self, Above);
} /* }}} */

/* SubtlextClientRestackLower {{{ */
static VALUE
SubtlextClientRestackLower(VALUE self)
{
  return SubtlextRestack(self, Below);
} /* }}} */

/* SubtlextClientMatchUp {{{ */
static VALUE
SubtlextClientMatchUp(VALUE self)
{
  return SubtlextMatch(self, SUB_WINDOW_UP);
} /* }}} */

/* SubtlextClientMatchLeft {{{ */
static VALUE
SubtlextClientMatchLeft(VALUE self)
{
  return SubtlextMatch(self, SUB_WINDOW_LEFT);
} /* }}} */

/* SubtlextClientMatchRight {{{ */
static VALUE
SubtlextClientMatchRight(VALUE self)
{
  return SubtlextMatch(self, SUB_WINDOW_RIGHT);
} /* }}} */

/* SubtlextClientMatchDown {{{ */
static VALUE
SubtlextClientMatchDown(VALUE self)
{
  return SubtlextMatch(self, SUB_WINDOW_DOWN);
} /* }}} */

/* SubtlextClientFocus {{{ */
static VALUE
SubtlextClientFocus(VALUE self)
{
  VALUE win = rb_iv_get(self, "@win");
  SubMessageData data = { { 0, 0, 0, 0, 0 } };

  if(RTEST(win))
    {
      data.l[0] = NUM2LONG(win);
      subSharedMessage(DefaultRootWindow(display), "_NET_ACTIVE_WINDOW", data, True);
    }
  else rb_raise(rb_eStandardError, "Failed setting focus");

  return Qnil;
} /* }}} */

/* SubtlextClientFocusHas {{{ */
static VALUE
SubtlextClientFocusHas(VALUE self)
{
  VALUE ret = Qfalse, win = Qnil;
  unsigned long *focus = NULL;

  /* Fetch data */
  win   = rb_iv_get(self, "@win");
  focus = (unsigned long *)subSharedPropertyGet(DefaultRootWindow(display),
    XA_WINDOW, "_NET_ACTIVE_WINDOW", NULL);

  if(RTEST(win) && focus)
    {
      if(*focus == NUM2LONG(win)) ret = Qtrue;
      free(focus);
    }

  return ret;
} /* }}} */

/* SubtlextClientKill {{{ */
static VALUE
SubtlextClientKill(VALUE self)
{
  VALUE win = rb_iv_get(self, "@win");
  SubMessageData data = { { 0, 0, 0, 0, 0 } };

  if(RTEST(win))
    {
      data.l[0] = CurrentTime;
      data.l[1] = 2; ///< Claim to be a pager

      subSharedMessage(NUM2LONG(win), "_NET_CLOSE_WINDOW", data, True);
    }
  else rb_raise(rb_eStandardError, "Failed killing client");

  return Qnil;
} /* }}} */

/* SubtlextClientToString {{{ */
static VALUE
SubtlextClientToString(VALUE self)
{
  VALUE name = rb_iv_get(self, "@name");

  return RTEST(name) ? name : Qnil;
} /* }}} */

/* SubtlextClientGravity {{{ */
static VALUE
SubtlextClientGravity(VALUE self)
{
  VALUE gravity = rb_iv_get(self, "@gravity");

  return RTEST(gravity) ? gravity : Qnil;
} /* }}} */

/* SubtlextClientGravitySet {{{ */
static VALUE
SubtlextClientGravitySet(VALUE self,
  VALUE value)
{
  if(T_FIXNUM == rb_type(value))
    {
      int gravity = FIX2INT(value);

      if(1 <= gravity && 9 >= gravity)
        {
          SubMessageData data = { { 0, 0, 0, 0, 0 } };
          VALUE id = rb_iv_get(self, "@id");

          data.l[0] = NUM2LONG(id);
          data.l[1] = FIX2LONG(value);

          subSharedMessage(DefaultRootWindow(display), "SUBTLE_WINDOW_GRAVITY", data, True);

          rb_iv_set(self, "@gravity", value);

          return value;
        }
      else
        {
          rb_raise(rb_eStandardError, "Failed setting client gravity");
          return Qnil;
        }
    }

  rb_raise(rb_eArgError, "Failed setting value type `%d'", rb_type(value));
  return Qnil;
} /* }}} */

/* SubtlextClientScreen {{{ */
static VALUE
SubtlextClientScreen(VALUE self)
{
  VALUE screen = rb_iv_get(self, "@screen");

  return SubtlextFind(SUB_TYPE_SCREEN, screen, False);
} /* }}} */

/* SubtlextClientScreenSet {{{ */
static VALUE
SubtlextClientScreenSet(VALUE self,
  VALUE value)
{
  int screen = -1;

  /* Check object type */
  switch(rb_type(value))
    {
      case T_FIXNUM: screen = FIX2INT(value); break;
      case T_OBJECT:
      case T_CLASS:
        {
          VALUE klass = rb_const_get(mod, rb_intern("Screen")); 
          
          if(rb_obj_is_instance_of(value, klass)) ///< Check object instance
            screen = FIX2INT(rb_iv_get(value, "@id"));
        }
    }

  /* Set screen */
  if(-1 != screen)
    {
      SubMessageData data = { { 0, 0, 0, 0, 0 } };
      VALUE id = rb_iv_get(self, "@id");

      data.l[0] = NUM2LONG(id);
      data.l[1] = screen;

      subSharedMessage(DefaultRootWindow(display), "SUBTLE_WINDOW_SCREEN", data, True);

      rb_iv_set(self, "@screen", INT2FIX(screen));

      return Qnil;
    }

  rb_raise(rb_eArgError, "Failed setting value type `%d'", rb_type(value));
  return Qnil;
} /* }}} */

/* SubtlextClientSize {{{ */
static VALUE
SubtlextClientSize(VALUE self)
{
  VALUE ary = rb_ary_new2(4);

  /* Push sizes */
  rb_ary_push(ary,  rb_iv_get(self, "@x"));
  rb_ary_push(ary,  rb_iv_get(self, "@y"));
  rb_ary_push(ary,  rb_iv_get(self, "@width"));
  rb_ary_push(ary,  rb_iv_get(self, "@height"));

  return ary;
} /* }}} */

/* SubtlextClientSizeSet {{{ */
static VALUE
SubtlextClientSizeSet(VALUE self,
  VALUE value)
{
  int len = 0;

  /* Check object */
  switch(rb_type(value))
    {
      case T_ARRAY:
        if(4 == (len = FIX2INT(rb_funcall(value, rb_intern("length"), 0, NULL))))
          {
            int i;
            VALUE win = rb_iv_get(self, "@win"), meth = rb_intern("at");
            SubMessageData data = { { 0, 0, 0, 0, 0 } };

            /* Get sizes */
            for(i = 1; i < 5; i++)
              data.l[i] = FIX2INT(rb_funcall(value, meth, 1, INT2FIX(i - 1)));

            /* Update geometry */
            rb_iv_set(self, "@x",      INT2FIX(data.l[1]));
            rb_iv_set(self, "@y",      INT2FIX(data.l[2]));
            rb_iv_set(self, "@width",  INT2FIX(data.l[3]));
            rb_iv_set(self, "@height", INT2FIX(data.l[4]));

            subSharedMessage(NUM2LONG(win), "_NET_MOVERESIZE_WINDOW", data, True);
          }
        else rb_raise(rb_eArgError, "Missing arguments");
        break;
      default: rb_raise(rb_eArgError, "Unknown value type");
    }

  return Qnil;
} /* }}} */

/* SubtlextClientUpdate {{{ */
VALUE
SubtlextClientUpdate(VALUE self)
{
   Window win = 0;

  /* Get real id */
  if((win = NUM2LONG(rb_iv_get(self, "@win"))))
    {
      int id = 0;
      char buf[20], *wmname = NULL, *wmklass = NULL;
      int *gravity = NULL, *screen = NULL, *flags = NULL;
      XWindowAttributes attrs;

      /* Get client id */
      snprintf(buf, sizeof(buf), "%#lx", win);
      id = subSharedClientFind(buf, NULL);

      /* Create new instance */
      subSharedPropertyClass(win, &wmname, &wmklass);
      gravity = (int *)subSharedPropertyGet(win, XA_CARDINAL,
        "SUBTLE_WINDOW_GRAVITY", NULL);
      screen  = (int *)subSharedPropertyGet(win, XA_CARDINAL,
        "SUBTLE_WINDOW_SCREEN", NULL);
      flags   = (int *)subSharedPropertyGet(win, XA_CARDINAL,
        "SUBTLE_WINDOW_FLAGS", NULL);

      XGetWindowAttributes(display, win, &attrs);

      /* Update properties */
      rb_iv_set(self, "@id",      INT2FIX(id));
      rb_iv_set(self, "@name",    rb_str_new2(wmname));
      rb_iv_set(self, "@klass",   rb_str_new2(wmklass));
      rb_iv_set(self, "@x",       INT2FIX(attrs.x));
      rb_iv_set(self, "@y",       INT2FIX(attrs.y));
      rb_iv_set(self, "@width",   INT2FIX(attrs.width));
      rb_iv_set(self, "@height",  INT2FIX(attrs.height));
      rb_iv_set(self, "@gravity", INT2FIX(*gravity));
      rb_iv_set(self, "@screen",  INT2FIX(*screen));
      rb_iv_set(self, "@flags",   INT2FIX(*flags));

      free(wmname);
      free(wmklass);
      free(gravity);
      free(screen);
      free(flags);
    }

  return Qnil;
} /* }}} */

/* SubtlextClientOperatorPlus {{{ */
static VALUE
SubtlextClientOperatorPlus(VALUE self,
  VALUE value)
{
  return SubtlextTag(self, value, SUB_ACTION_TAG, SUB_TYPE_CLIENT);
} /* }}} */

/* SubtlextClientOperatorMinus {{{ */
static VALUE
SubtlextClientOperatorMinus(VALUE self,
  VALUE value)
{
  return SubtlextTag(self, value, SUB_ACTION_UNTAG, SUB_TYPE_CLIENT);
} /* }}} */

/* SubtlextScreenInit {{{ */
static VALUE
SubtlextScreenInit(VALUE self,
  VALUE id)
{
  rb_iv_set(self, "@id",      id);
  rb_iv_set(self, "@x",       Qnil);
  rb_iv_set(self, "@y",       Qnil); 
  rb_iv_set(self, "@width",   Qnil); 
  rb_iv_set(self, "@height",  Qnil); 

  return self;
} /* }}} */

/* SubtlextScreenClientList {{{ */
static VALUE
SubtlextScreenClientList(VALUE self)
{
  int i, id, size = 0;
  Window *clients = NULL;
  VALUE array = Qnil;
  
  id      = FIX2INT(rb_iv_get(self, "@id"));
  array   = rb_ary_new();
  clients = subSharedClientList(&size);

  if(clients)
    {
      for(i = 0; i < size; i++)
        {
          int *screen = (int *)subSharedPropertyGet(clients[i], XA_CARDINAL, 
            "SUBTLE_WINDOW_SCREEN", NULL);

          if(id - 1 == *screen) ///< Check if screen matches
            rb_ary_push(array, SubtlextClientFromWin(clients[i]));

          free(screen);
        }
      free(clients);
    }

  return array;
} /* }}} */

/* SubtlextSubtleKill {{{ */
static void
SubtlextSubtleKill(void *data)
{
  if(0 == --refcount) 
    {
      XCloseDisplay(display);
      display = NULL;
      subSharedLogDebug("Connection closed\n");
    }
} /* }}} */

/* SubtlextSubtleNew {{{ */
static VALUE
SubtlextSubtleNew(int argc,
  VALUE *argv,
  VALUE self)
{
  char *name = NULL;
  VALUE disp = Qnil, data = Qnil;

#ifdef DEBUG
  VALUE dbg = Qfalse;

  rb_scan_args(argc, argv, "02", &disp, &dbg);

  if(Qtrue == dbg) debug++;
#else
  rb_scan_args(argc, argv, "01", &disp);
#endif /* DEBUG */ 

  /* Open display */
  if(!display) ///< Establish new connection
    {
      if(RTEST(disp)) name = RSTRING_PTR(disp);
      if(!(display = XOpenDisplay(name)))
        {
          subSharedLogError("Failed opening display `%s'\n", (name) ? name : ":0.0");

          return Qnil;
        }
      XSetErrorHandler(subSharedLogXError);

      /* Check if subtle is running */
      if(True != subSharedSubtleRunning())
        {
          XCloseDisplay(display);
          display = NULL;
          
          rb_raise(rb_eStandardError, "Failed finding running %s", PKG_NAME);

          return Qnil;
        }

      subSharedLogDebug("Connection opened (%s)\n", name);
    }
  refcount++;

  /* Create instance */
  data = Data_Wrap_Struct(self, 0, SubtlextSubtleKill, display);
  rb_obj_call_init(data, 0, NULL);

  return data;
} /* }}} */

/* SubtlextSubtleNew2 {{{ */
static VALUE
SubtlextSubtleNew2(VALUE self,
  VALUE disp)
{
  VALUE data = Qnil;

  display = (Display *)disp;
  refcount++;

  /* Create instance */
  data = Data_Wrap_Struct(self, 0, NULL, display);
  rb_obj_call_init(data, 0, NULL);

  return data;
} /* }}} */

/* SubtlextSubtleVersion {{{ */
VALUE 
SubtlextSubtleVersion(VALUE self)
{
  return rb_str_new2(PKG_VERSION);
} /* }}} */

/* SubtlextSubtleDisplay {{{ */
VALUE 
SubtlextSubtleDisplay(VALUE self)
{
  return rb_str_new2(DisplayString(display));
} /* }}} */

/* SubtlextSubtleRunning {{{ */
static VALUE
SubtlextSubtleRunning(VALUE self)
{
  return subSharedSubtleRunning() ? Qtrue : Qfalse;
} /* }}} */

/* SubtlextSubtleReload {{{ */
static VALUE
SubtlextSubtleReload(VALUE self)
{
  SubMessageData data = { { 0, 0, 0, 0, 0 } };

  subSharedMessage(DefaultRootWindow(display), "SUBTLE_RELOAD", data, True);

  return Qnil;
} /* }}} */

/* SubtlextSubtleQuit {{{ */
static VALUE
SubtlextSubtleQuit(VALUE self)
{
  SubMessageData data = { { 0, 0, 0, 0, 0 } };

  subSharedMessage(DefaultRootWindow(display), "SUBTLE_QUIT", data, True);

  return Qnil;
} /* }}} */

/* SubtlextSubtleTagList {{{ */
static VALUE
SubtlextSubtleTagList(VALUE self)
{
  int i, size = 0;
  char **tags = NULL;
  VALUE array = Qnil, method = Qnil, klass = Qnil;
  
  method = rb_intern("new");
  klass  = rb_const_get(mod, rb_intern("Tag"));
  tags   = subSharedPropertyStrings(DefaultRootWindow(display), "SUBTLE_TAG_LIST", &size);
  array  = rb_ary_new2(size);

  for(i = 0; i < size; i++)
    {
      VALUE t = rb_funcall(klass, method, 1, rb_str_new2(tags[i]));

      rb_iv_set(t, "@id",  INT2FIX(i));
      rb_ary_push(array, t);
    }

  XFreeStringList(tags);

  return array;
} /* }}} */

/* SubtlextSubtleTagFind {{{ */
static VALUE
SubtlextSubtleTagFind(VALUE self,
  VALUE name)
{
  return SubtlextFind(SUB_TYPE_TAG, name, False);
} /* }}} */

/* SubtlextSubtleTagAdd {{{ */
static VALUE
SubtlextSubtleTagAdd(VALUE self,
  VALUE value)
{
  return SubtlextFind(SUB_TYPE_TAG, value, True); ///< Find or create tag
} /* }}} */

/* SubtlextSubtleTagDel {{{ */
static VALUE
SubtlextSubtleTagDel(VALUE self,
  VALUE value)
{
  VALUE tag = Qnil;

  if(Qnil != (tag = SubtlextFind(SUB_TYPE_TAG, value, False)))
    {
      VALUE id = Qnil;
      SubMessageData data = { { 0, 0, 0, 0, 0 } };

      id = rb_iv_get(tag, "@id");

      /* Send data */
      data.l[0] = FIX2LONG(id);

      subSharedMessage(DefaultRootWindow(display), "SUBTLE_TAG_KILL", data, True);  
    }
  else rb_raise(rb_eStandardError, "Failed finding tag");

  return Qnil;
} /* }}} */

/* SubtlextSubtleClientList {{{ */
static VALUE
SubtlextSubtleClientList(VALUE self)
{
  int size = 0;
  Window *clients = NULL;
  VALUE array = Qnil;

  array  = rb_ary_new2(size);    

  if((clients = subSharedClientList(&size)))
    {
      int i;

      for(i = 0; i < size; i++)
        rb_ary_push(array, SubtlextClientFromWin(clients[i]));

      free(clients);
    }

  return array;
} /* }}} */

/* SubtlextSubtleClientFind {{{ */
static VALUE
SubtlextSubtleClientFind(VALUE self,
  VALUE name)
{
  int id = -1;
  Window win = 0;
  VALUE client = Qnil;

  if(RTEST(name) && -1 != ((id = subSharedClientFind(RSTRING_PTR(name), &win))))
    {
      client = SubtlextClientFromWin(win);
    }
  else rb_raise(rb_eStandardError, "Failed finding client");

  return client ;
} /* }}} */

/* SubtlextSubtleClientFocus {{{ */
static VALUE
SubtlextSubtleClientFocus(VALUE self,
  VALUE name)
{
  int id = -1;
  Window win = 0;
  VALUE client = Qnil;

  if(RTEST(name) && -1 != ((id = subSharedClientFind(RSTRING_PTR(name), &win))))
    {
      SubMessageData data = { { 0, 0, 0, 0, 0 } };

      data.l[0] = win;
      subSharedMessage(DefaultRootWindow(display), "_NET_ACTIVE_WINDOW", data, True);

      client = SubtlextClientFromWin(win);
    }
  else rb_raise(rb_eStandardError, "Failed setting focus");

  return client;
} /* }}} */

/* SubtlextSubtleClientFocusLeft {{{ */
static VALUE
SubtlextSubtleClientFocusLeft(VALUE self)
{
  return SubtlextClientFocus(SubtlextClientMatchLeft(SubtlextSubtleClientCurrent(self)));
} /* }}} */

/* SubtlextSubtleClientFocusDown {{{ */
static VALUE
SubtlextSubtleClientFocusDown(VALUE self)
{
  return SubtlextClientFocus(SubtlextClientMatchDown(SubtlextSubtleClientCurrent(self)));
} /* }}} */

/* SubtlextSubtleClientFocusUp {{{ */
static VALUE
SubtlextSubtleClientFocusUp(VALUE self)
{
  return SubtlextClientFocus(SubtlextClientMatchUp(SubtlextSubtleClientCurrent(self)));
} /* }}} */

/* SubtlextSubtleClientFocusRight {{{ */
static VALUE
SubtlextSubtleClientFocusRight(VALUE self)
{
  return SubtlextClientFocus(SubtlextClientMatchRight(SubtlextSubtleClientCurrent(self)));
} /* }}} */

/* SubtlextSubtleClientDel {{{ */
static VALUE
SubtlextSubtleClientDel(VALUE self,
  VALUE name)
{
  int id = -1;
  Window win = 0;

  if(RTEST(name) && -1 != ((id = subSharedClientFind(RSTRING_PTR(name), &win))))
    {
      SubMessageData data = { { 0, 0, 0, 0, 0 } };

      /* Send data */
      data.l[0] = CurrentTime;
      data.l[1] = 2; ///< Claim to be a pager

      subSharedMessage(win, "_NET_CLOSE_WINDOW", data, True);       
    }
  else rb_raise(rb_eStandardError, "Failed finding client");

  return Qnil;
} /* }}} */

/* SubtlextSubtleClientCurrent {{{ */
static VALUE
SubtlextSubtleClientCurrent(VALUE self)
{
  VALUE client = Qnil;
  unsigned long *focus = NULL;

  if((focus = (unsigned long *)subSharedPropertyGet(DefaultRootWindow(display),
    XA_WINDOW, "_NET_ACTIVE_WINDOW", NULL)))
    {
      client = SubtlextClientFromWin(*focus);

      free(focus);
    }
  else rb_raise(rb_eStandardError, "Failed getting current client");

  return client;
} /* }}} */

/* SubtlextSubtleSubletDel {{{ */
static VALUE
SubtlextSubtleSubletDel(VALUE self,
  VALUE name)
{
  int id = -1;

  if(RTEST(name) && -1 != ((id = subSharedSubletFind(RSTRING_PTR(name)))))
    {
      SubMessageData data = { { 0, 0, 0, 0, 0 } };

      /* Send data */
      data.l[0] = id;

      subSharedMessage(DefaultRootWindow(display), "SUBTLE_SUBLET_KILL", data, True);       
    }
  else rb_raise(rb_eStandardError, "Failed finding sublet");

  return Qnil;
} /* }}} */

/* SubtlextSubtleSubletList {{{ */
static VALUE
SubtlextSubtleSubletList(VALUE self)
{
  int i, size = 0;
  char **sublets = NULL;
  VALUE array = Qnil, method = Qnil, klass = Qnil;
  
  method  = rb_intern("new");
  klass   = rb_const_get(mod, rb_intern("Sublet"));
  sublets = subSharedPropertyStrings(DefaultRootWindow(display), "SUBTLE_SUBLET_LIST", &size);
  array   = rb_ary_new2(size);

  if(sublets)
    {
      for(i = 0; i < size; i++)
        {
          VALUE s = rb_funcall(klass, method, 1, rb_str_new2(sublets[i]));

          rb_iv_set(s, "@id", INT2FIX(i));
          rb_ary_push(array, s);
        }

      XFreeStringList(sublets);
    }

  return array;
} /* }}} */

/* SubtlextSubtleSubletFind {{{ */
static VALUE
SubtlextSubtleSubletFind(VALUE self,
  VALUE name)
{
  int id = -1;
  VALUE klass = Qnil, sublet = Qnil;

  if(RTEST(name) && -1 != ((id = subSharedSubletFind(RSTRING_PTR(name)))))
    {
      klass  = rb_const_get(mod, rb_intern("Sublet"));
      sublet = rb_funcall(klass, rb_intern("new"), 1, name);

      rb_iv_set(sublet, "@id", INT2FIX(id));

      return sublet;
    }
  else rb_raise(rb_eStandardError, "Failed finding sublet");

  return sublet;
} /* }}} */

/* SubtlextSubtleViewList {{{ */
static VALUE
SubtlextSubtleViewList(VALUE self)
{
  int i, size = 0;
  char **names = NULL;
  Window *views = NULL;
  VALUE array = Qnil, method = Qnil, klass = Qnil;
  
  /* Collect data */
  method = rb_intern("new");
  klass  = rb_const_get(mod, rb_intern("View"));
  names  = subSharedPropertyStrings(DefaultRootWindow(display), 
    "_NET_DESKTOP_NAMES", &size);
  views   = (Window *)subSharedPropertyGet(DefaultRootWindow(display), 
    XA_WINDOW, "_NET_VIRTUAL_ROOTS", NULL);  
  array  = rb_ary_new2(size);

  if(names && views)
    {
      for(i = 0; i < size; i++)
        {
          VALUE v = rb_funcall(klass, method, 1, rb_str_new2(names[i]));

          rb_iv_set(v, "@id",  INT2FIX(i));
          rb_iv_set(v, "@win", LONG2NUM(views[i]));
          rb_ary_push(array, v);
        }

      XFreeStringList(names); 
      free(views);
    }

  return array;
} /* }}} */

/* SubtlextSubtleViewFind {{{ */
static VALUE
SubtlextSubtleViewFind(VALUE self,
  VALUE name)
{
  return SubtlextFind(SUB_TYPE_VIEW, name, False);
} /* }}} */

/* SubtlextSubtleViewAdd {{{ */
static VALUE
SubtlextSubtleViewAdd(VALUE self,
  VALUE value)
{
  return SubtlextFind(SUB_TYPE_VIEW, value, True); ///< Find or create view
} /* }}} */

/* SubtlextSubtleViewDel {{{ */
static VALUE
SubtlextSubtleViewDel(VALUE self,
  VALUE value)
{
  VALUE view = Qnil;

  if(Qnil != (view = SubtlextFind(SUB_TYPE_VIEW, value, False)))
    {
      VALUE id = Qnil;
      SubMessageData data = { { 0, 0, 0, 0, 0 } };

      id = rb_iv_get(view, "@id");

      /* Send data */
      data.l[0] = FIX2LONG(id);

      subSharedMessage(DefaultRootWindow(display), "SUBTLE_VIEW_KILL", data, True);  
    }

  return Qnil;
} /* }}} */

/* SubtlextSubtleViewCurrent {{{ */
static VALUE
SubtlextSubtleViewCurrent(VALUE self)
{
  int size = 0;
  char **names = NULL;
  unsigned long *cv = NULL;
  Window *views = NULL;
  VALUE klass = Qnil, view = Qnil;
  
  /* Create view instance */
  klass = rb_const_get(mod, rb_intern("View"));
  names = subSharedPropertyStrings(DefaultRootWindow(display), "_NET_DESKTOP_NAMES", &size);
  cv    = (unsigned long *)subSharedPropertyGet(DefaultRootWindow(display),
    XA_CARDINAL, "_NET_CURRENT_DESKTOP", NULL);
  views = (Window *)subSharedPropertyGet(DefaultRootWindow(display), XA_WINDOW, 
    "_NET_VIRTUAL_ROOTS", NULL);
  view  = rb_funcall(klass, rb_intern("new"), 1, rb_str_new2(names[*cv]));

  rb_iv_set(view, "@id",  INT2FIX(*cv));
  rb_iv_set(view, "@win", LONG2NUM(views[*cv]));

  XFreeStringList(names);
  free(views);
  free(cv);
      
  return view;
} /* }}} */

/* SubtlextSubtleScreenList {{{ */
static VALUE
SubtlextSubtleScreenList(VALUE self)
{
  return SubtlextScreens();
} /* }}} */

/* SubtlextSubtleScreenCurrent {{{ */
static VALUE
SubtlextSubtleScreenCurrent(VALUE self)
{
  int num = 0;
  unsigned long *focus = NULL;

  /* Get current screen from current client or use the first */
  if((focus = (unsigned long *)subSharedPropertyGet(DefaultRootWindow(display),
    XA_WINDOW, "_NET_ACTIVE_WINDOW", NULL)))
    {
      int *screen = NULL;

      screen = (int *)subSharedPropertyGet(*focus, XA_CARDINAL, "SUBTLE_WINDOW_SCREEN", NULL);
      num    = *screen;

      free(focus);
      free(screen);
    }

  return SubtlextFind(SUB_TYPE_SCREEN, INT2FIX(num), False);
} /* }}} */

/* SubtlextSubtleScreenFind {{{ */
static VALUE
SubtlextSubtleScreenFind(VALUE self,
  VALUE id)
{
  return SubtlextFind(SUB_TYPE_SCREEN, id, False);
} /* }}} */

/* SubtlextScreenToString {{{ */
static VALUE
SubtlextScreenToString(VALUE self)
{
  char buf[256];
  int x = 0, y = 0, width = 0, height = 0;

  x      = FIX2INT(rb_iv_get(self, "@x"));
  y      = FIX2INT(rb_iv_get(self, "@y"));
  width  = FIX2INT(rb_iv_get(self, "@width"));
  height = FIX2INT(rb_iv_get(self, "@height"));

  snprintf(buf, sizeof(buf), "%dx%d+%d+%d", x, y, width, height);

  return rb_str_new2(buf);
} /* }}} */

/* SubtlextSubtleToString {{{ */
static VALUE
SubtlextSubtleToString(VALUE self)
{
  char buf[256];

  snprintf(buf, sizeof(buf), "%s (%s) on %s (%dx%d)", 
    PKG_NAME, PKG_VERSION, DisplayString(display), 
    DisplayWidth(display, DefaultScreen(display)),
    DisplayHeight(display, DefaultScreen(display)));

  return rb_str_new2(buf);
} /* }}} */

/* SubtlextSubletInit {{{ */
static VALUE
SubtlextSubletInit(VALUE self,
  VALUE name)
{
  rb_iv_set(self, "@id",   Qnil);
  rb_iv_set(self, "@name", name);

  return self;
} /* }}} */

/* SubtlextSubletUpdate {{{ */
static VALUE
SubtlextSubletUpdate(VALUE self)
{
  int id = -1;
  VALUE name = rb_iv_get(self, "@name");

  if(RTEST(name) && -1 != ((id = subSharedSubletFind(RSTRING_PTR(name)))))
    {
      SubMessageData data = { { 0, 0, 0, 0, 0 } };

      /* Send data */
      data.l[0] = id;

      subSharedMessage(DefaultRootWindow(display), "SUBTLE_SUBLET_UPDATE", data, True);
    }
  else rb_raise(rb_eStandardError, "Failed finding sublet");

  return Qnil;
} /* }}} */

/* SubtlextDataScreen {{{ */
static VALUE
SubtlextSubletData(VALUE self)
{
  VALUE id = rb_iv_get(self, "@id");

  //gravity = (int *)subSharedPropertyGet(win, XA_STRING,
    //"SUBTLE_SUBLET_DATA", NULL);
   
      
  return RTEST(id) ? id : Qnil;
} /* }}} */

/* SubtlextSubletDataSet {{{ */
static VALUE
SubtlextSubletDataSet(VALUE self,
  VALUE value)
{
  if(T_STRING == rb_type(value))
    {
      SubMessageData data = { { 0, 0, 0, 0, 0 } };
      VALUE id = rb_iv_get(self, "@id");

      snprintf(data.b, sizeof(data.b), "%c%s",
        (char)NUM2LONG(id), RSTRING_PTR(value));

      subSharedMessage(DefaultRootWindow(display), "SUBTLE_SUBLET_DATA", data, True);

      return value;
    }
  else
    {
      rb_raise(rb_eStandardError, "Failed setting sublet data");
      return Qnil;
    }

  rb_raise(rb_eArgError, "Failed setting value type `%d'", rb_type(value));
  return Qnil;
} /* }}} */

/* SubtlextSubletToString {{{ */
static VALUE
SubtlextSubletToString(VALUE self)
{
  VALUE name = rb_iv_get(self, "@name");

  return RTEST(name) ? name : Qnil;
} /* }}} */

/* SubtlextTagInit {{{ */
static VALUE
SubtlextTagInit(VALUE self,
  VALUE name)
{
  rb_iv_set(self, "@id",   Qnil);
  rb_iv_set(self, "@name", name);

  return self;
} /* }}} */

/* SubtlextTagSave {{{ */
static VALUE
SubtlextTagSave(VALUE self)
{
  return SubtlextFind(SUB_TYPE_TAG, self, True);
} /* }}} */

/* SubtlextTagTagggings {{{ */
static VALUE
SubtlextTagTaggings(VALUE self)
{
  int i, id = 0, size_clients = 0, size_views = 0;
  char **names = NULL;
  Window *clients = NULL, *views = NULL;
  VALUE array = Qnil, method = Qnil, klass = Qnil;
  
  /* Collect data */
  id      = FIX2INT(rb_iv_get(self, "@id"));
  method  = rb_intern("new");
  array   = rb_ary_new2(size_clients);
  clients = subSharedClientList(&size_clients);
  names   = subSharedPropertyStrings(DefaultRootWindow(display), 
    "_NET_DESKTOP_NAMES", &size_views);
  views   = (Window *)subSharedPropertyGet(DefaultRootWindow(display), 
    XA_WINDOW, "_NET_VIRTUAL_ROOTS", NULL);

  /* Clients */
  if(clients)
    {
      klass = rb_const_get(mod, rb_intern("Client"));

      for(i = 0; i < size_clients; i++)
        {
          unsigned long *flags = (unsigned long *)subSharedPropertyGet(clients[i], XA_CARDINAL, 
            "SUBTLE_WINDOW_TAGS", NULL);

          if((int)*flags & (1L << (id + 1))) ///< Check if tag id matches
            {
              VALUE c = Qnil;
              char *wmklass = NULL;
              
              /* Create client instance */
              subSharedPropertyClass(clients[i], NULL, &wmklass);
              c = rb_funcall(klass, method, 1, rb_str_new2(wmklass));

              rb_iv_set(c, "@id",  INT2FIX(i));
              rb_iv_set(c, "@win", LONG2NUM(clients[i]));
              rb_ary_push(array, c);

              free(wmklass);
            }

          free(flags);
        }
      free(clients);
    }

  /* Views */
  if(names && views)
    {
      klass = rb_const_get(mod, rb_intern("View"));

      for(i = 0; i < size_views; i++)
        {
          unsigned long *flags = (unsigned long *)subSharedPropertyGet(views[i], XA_CARDINAL, 
            "SUBTLE_WINDOW_TAGS", NULL);

          if((int)*flags & (1L << (id + 1))) ///< Check if tag id matches
            {
              VALUE v = rb_funcall(klass, method, 1, rb_str_new2(names[i]));

              rb_iv_set(v, "@id",  INT2FIX(i));
              rb_iv_set(v, "@win", LONG2NUM(views[i]));
              rb_ary_push(array, v);
            }

          free(flags);
        }

      XFreeStringList(names);
      free(views);
    }

  return array;
} /* }}} */

/* SubtlextTagToString {{{ */
static VALUE
SubtlextTagToString(VALUE self)
{
  VALUE name = rb_iv_get(self, "@name");

  return RTEST(name) ? name : Qnil;
} /* }}} */

/* SubtlextViewInit {{{ */
static VALUE
SubtlextViewInit(VALUE self,
  VALUE name)
{
  rb_iv_set(self, "@id",   Qnil);
  rb_iv_set(self, "@win",  Qnil);
  rb_iv_set(self, "@name", name);

  return self;
} /* }}} */

/* SubtlextViewClientList {{{ */
static VALUE
SubtlextViewClientList(VALUE self)
{
  int i, size = 0;
  Window *clients = NULL;
  VALUE win = Qnil, array = Qnil;
  unsigned long *flags1 = NULL;
  
  win     = rb_iv_get(self, "@win");
  array   = rb_ary_new2(size);
  clients = subSharedClientList(&size);
  flags1  = (unsigned long *)subSharedPropertyGet(NUM2LONG(win), XA_CARDINAL, 
    "SUBTLE_WINDOW_TAGS", NULL);

  if(clients)
    {
      for(i = 0; i < size; i++)
        {
          unsigned long *flags2 = (unsigned long *)subSharedPropertyGet(clients[i], XA_CARDINAL, 
            "SUBTLE_WINDOW_TAGS", NULL);

          if(*flags1 & *flags2) ///< Check if there are common tags
            rb_ary_push(array, SubtlextClientFromWin(clients[i]));

          free(flags2);
        }
      free(clients);
    }

  return array;
} /* }}} */

/* SubtlextViewTagList {{{ */
static VALUE
SubtlextViewTagList(VALUE self)
{
  Window win = 0;
  int i, size = 0;
  char **tags = NULL;
  unsigned long *flags = NULL;
  VALUE array = Qnil, method = Qnil, klass = Qnil, name = Qnil;

  name = rb_iv_get(self, "@name");
  if(RTEST(name) && -1 != subSharedViewFind(RSTRING_PTR(name), &win))
    {
      method = rb_intern("new");
      klass  = rb_const_get(mod, rb_intern("Tag"));
      flags  = (unsigned long *)subSharedPropertyGet(win, XA_CARDINAL, 
        "SUBTLE_WINDOW_TAGS", NULL);
      tags   = subSharedPropertyStrings(DefaultRootWindow(display), "SUBTLE_TAG_LIST", &size);
      array  = rb_ary_new2(size);

      /* Build tag array */
      for(i = 0; i < size; i++)
        {
          if((int)*flags & (1L << (i + 1)))
            {
              VALUE t = rb_funcall(klass, method, 1, rb_str_new2(tags[i]));

              rb_iv_set(t, "@id", INT2FIX(i));
              rb_ary_push(array, t);
            }
        }

      XFreeStringList(tags);

      return array;
    }
  
  return Qnil;
} /* }}} */

/* SubtlextViewTagAdd {{{ */
static VALUE
SubtlextViewTagAdd(VALUE self,
  VALUE value)
{
  return SubtlextTag(self, value, SUB_ACTION_TAG, SUB_TYPE_VIEW);
} /* }}} */

/* SubtlextViewTagDel {{{ */
static VALUE
SubtlextViewTagDel(VALUE self,
  VALUE value)
{  
  return SubtlextTag(self, value, SUB_ACTION_UNTAG, SUB_TYPE_VIEW);
} /* }}} */

/* SubtlextViewJump {{{ */
static VALUE
SubtlextViewJump(VALUE self)
{
  VALUE id = rb_iv_get(self, "@id");
  SubMessageData data = { { 0, 0, 0, 0, 0 } };

  data.l[0] = FIX2INT(id);

  subSharedMessage(DefaultRootWindow(display), "_NET_CURRENT_DESKTOP", data, True);

  return Qnil;
} /* }}} */

/* SubtlextViewCurrent {{{ */
static VALUE
SubtlextViewCurrent(VALUE self)
{
  unsigned long *cv = NULL;
  VALUE id = Qnil, ret = Qfalse;;
  
  id = rb_iv_get(self, "@id");
  cv = (unsigned long *)subSharedPropertyGet(DefaultRootWindow(display),
    XA_CARDINAL, "_NET_CURRENT_DESKTOP", NULL);

  if(FIX2INT(id) == *cv) ret = Qtrue;
  free(cv);
      
  return ret;
} /* }}} */

/* SubtlextViewSave {{{ */
static VALUE
SubtlextViewSave(VALUE self)
{
  return SubtlextFind(SUB_TYPE_VIEW, self, True);
} /* }}} */

/* SubtlextViewToString {{{ */
static VALUE
SubtlextViewToString(VALUE self)
{
  VALUE name = rb_iv_get(self, "@name");

  return RTEST(name) ? name : Qnil;
} /* }}} */

/* SubtlextViewOperatorPlus {{{ */
static VALUE
SubtlextViewOperatorPlus(VALUE self,
  VALUE value)
{
  return SubtlextTag(self, value, SUB_ACTION_TAG, SUB_TYPE_VIEW);
} /* }}} */

/* SubtlextViewOperatorMinus {{{ */
static VALUE
SubtlextViewOperatorMinus(VALUE self,
  VALUE value)
{
  return SubtlextTag(self, value, SUB_ACTION_UNTAG, SUB_TYPE_VIEW);
} /* }}} */

 /** Init_subtlext {{{
  * @brief Extension entry point
  **/

void 
Init_subtlext(void)
{
  VALUE klass = Qnil;

  /* Module: sublext */
  mod = rb_define_module("Subtlext");

  /* Class: client */
  klass = rb_define_class_under(mod, "Client", rb_cObject);
  rb_define_attr(klass,   "id",      1, 0);
  rb_define_attr(klass,   "win",     1, 0);
  rb_define_attr(klass,   "klass",   1, 0);
  rb_define_attr(klass,   "x",       1, 0);
  rb_define_attr(klass,   "y",       1, 0);
  rb_define_attr(klass,   "width",   1, 0);
  rb_define_attr(klass,   "height",  1, 0);
  rb_define_attr(klass,   "name",    1, 0);
  rb_define_attr(klass,   "flags",   1, 0);

  rb_define_method(klass, "initialize",   SubtlextClientInit,          1);
  rb_define_method(klass, "views",        SubtlextClientViewList,      0);
  rb_define_method(klass, "tags",         SubtlextClientTagList,       0);
  rb_define_method(klass, "tag",          SubtlextClientTagAdd,        1);
  rb_define_method(klass, "untag",        SubtlextClientTagDel,        1);
  rb_define_method(klass, "toggle_full",  SubtlextClientToggleFull,    0);
  rb_define_method(klass, "toggle_float", SubtlextClientToggleFloat,   0);
  rb_define_method(klass, "toggle_stick", SubtlextClientToggleStick,   0);
  rb_define_method(klass, "is_full?",     SubtlextClientStateFull,     0);
  rb_define_method(klass, "is_float?",    SubtlextClientStateFloat,    0);
  rb_define_method(klass, "is_stick?",    SubtlextClientStateStick,    0);
  rb_define_method(klass, "raise",        SubtlextClientRestackRaise,  0);
  rb_define_method(klass, "lower",        SubtlextClientRestackLower,  0);
  rb_define_method(klass, "up",           SubtlextClientMatchUp,       0);
  rb_define_method(klass, "left",         SubtlextClientMatchLeft,     0);
  rb_define_method(klass, "right",        SubtlextClientMatchRight,    0);
  rb_define_method(klass, "down",         SubtlextClientMatchDown,     0);
  rb_define_method(klass, "focus",        SubtlextClientFocus,         0);
  rb_define_method(klass, "focus?",       SubtlextClientFocusHas,      0);
  rb_define_method(klass, "kill",         SubtlextClientKill,          0);
  rb_define_method(klass, "to_str",       SubtlextClientToString,      0);
  rb_define_method(klass, "gravity",      SubtlextClientGravity,       0);
  rb_define_method(klass, "gravity=",     SubtlextClientGravitySet,    1);
  rb_define_method(klass, "screen",       SubtlextClientScreen,        0);
  rb_define_method(klass, "screen=",      SubtlextClientScreenSet,     1);  
  rb_define_method(klass, "size",         SubtlextClientSize,          0);
  rb_define_method(klass, "size=",        SubtlextClientSizeSet,       1);
  rb_define_method(klass, "update",       SubtlextClientUpdate,        0);
  rb_define_method(klass, "+",            SubtlextClientOperatorPlus,  1);
  rb_define_method(klass, "-",            SubtlextClientOperatorMinus, 1);
  rb_define_alias(klass, "to_s", "to_str");

  /* Class: gravity */
  klass = rb_define_class_under(mod, "Gravity", rb_cObject);
  rb_define_const(klass, "TopLeft",     INT2FIX(SUB_GRAVITY_TOP_LEFT));
  rb_define_const(klass, "Top",         INT2FIX(SUB_GRAVITY_TOP));
  rb_define_const(klass, "TopRight",    INT2FIX(SUB_GRAVITY_TOP_RIGHT));
  rb_define_const(klass, "Left",        INT2FIX(SUB_GRAVITY_LEFT));
  rb_define_const(klass, "Center",      INT2FIX(SUB_GRAVITY_CENTER));
  rb_define_const(klass, "Right",       INT2FIX(SUB_GRAVITY_RIGHT));
  rb_define_const(klass, "BottomLeft",  INT2FIX(SUB_GRAVITY_BOTTOM_LEFT));
  rb_define_const(klass, "Bottom",      INT2FIX(SUB_GRAVITY_BOTTOM));
  rb_define_const(klass, "BottomRight", INT2FIX(SUB_GRAVITY_BOTTOM_RIGHT));

  /* Class: screen */
  klass = rb_define_class_under(mod, "Screen", rb_cObject);
  rb_define_attr(klass,   "id",      1, 0);
  rb_define_attr(klass,   "x",       1, 0);
  rb_define_attr(klass,   "y",       1, 0);
  rb_define_attr(klass,   "width",   1, 0);
  rb_define_attr(klass,   "height",  1, 0);

  rb_define_method(klass, "initialize",   SubtlextScreenInit,       1);
  rb_define_method(klass, "clients",      SubtlextScreenClientList, 0);
  rb_define_method(klass, "to_str",       SubtlextScreenToString,   0);
  rb_define_alias(klass, "to_s", "to_str");

  /* Class: subtle */
  klass = rb_define_class_under(mod, "Subtle", rb_cObject);
  rb_define_singleton_method(klass, "new", SubtlextSubtleNew,  -1);
  rb_define_singleton_method(klass, "new2", SubtlextSubtleNew2, 1);

  rb_define_method(klass, "version",        SubtlextSubtleVersion,          0);
  rb_define_method(klass, "display",        SubtlextSubtleDisplay,          0);
  rb_define_method(klass, "views",          SubtlextSubtleViewList,         0);
  rb_define_method(klass, "tags",           SubtlextSubtleTagList,          0);
  rb_define_method(klass, "screens",        SubtlextSubtleScreenList,       0);
  rb_define_method(klass, "sublets",        SubtlextSubtleSubletList,       0);
  rb_define_method(klass, "clients",        SubtlextSubtleClientList,       0);
  rb_define_method(klass, "find_view",      SubtlextSubtleViewFind,         1);
  rb_define_method(klass, "find_tag",       SubtlextSubtleTagFind,          1);
  rb_define_method(klass, "find_client",    SubtlextSubtleClientFind,       1);
  rb_define_method(klass, "find_screen",    SubtlextSubtleScreenFind,       1);
  rb_define_method(klass, "find_sublet",    SubtlextSubtleSubletFind,       1);
  rb_define_method(klass, "focus_client",   SubtlextSubtleClientFocus,      1);
  rb_define_method(klass, "focus_left",     SubtlextSubtleClientFocusLeft,  0);
  rb_define_method(klass, "focus_down",     SubtlextSubtleClientFocusDown,  0);
  rb_define_method(klass, "focus_up",       SubtlextSubtleClientFocusUp,    0);
  rb_define_method(klass, "focus_right",    SubtlextSubtleClientFocusRight, 0);
  rb_define_method(klass, "add_tag",        SubtlextSubtleTagAdd,           1);
  rb_define_method(klass, "add_view",       SubtlextSubtleViewAdd,          1);
  rb_define_method(klass, "del_client",     SubtlextSubtleClientDel,        1);
  rb_define_method(klass, "del_sublet",     SubtlextSubtleSubletDel,        1);
  rb_define_method(klass, "del_tag",        SubtlextSubtleTagDel,           1);
  rb_define_method(klass, "del_view",       SubtlextSubtleViewDel,          1);
  rb_define_method(klass, "current_view",   SubtlextSubtleViewCurrent,      0);
  rb_define_method(klass, "current_client", SubtlextSubtleClientCurrent,    0);
  rb_define_method(klass, "current_screen", SubtlextSubtleScreenCurrent,    0);
  rb_define_method(klass, "running?",       SubtlextSubtleRunning,          0);
  rb_define_method(klass, "reload",         SubtlextSubtleReload,           0);
  rb_define_method(klass, "quit",           SubtlextSubtleQuit,             0);
  rb_define_method(klass, "to_str",         SubtlextSubtleToString,         0);
  rb_define_alias(klass, "to_s", "to_str");

  /* Class: sublet */
  klass = rb_define_class_under(mod, "Sublet", rb_cObject);
  rb_define_attr(klass,   "name", 1, 0);

  rb_define_method(klass, "initialize", SubtlextSubletInit,     1);
  rb_define_method(klass, "update",     SubtlextSubletUpdate,   0);
  rb_define_method(klass, "data",       SubtlextSubletData,     0);
  rb_define_method(klass, "data=",      SubtlextSubletDataSet,  1);  
  rb_define_method(klass, "to_s",       SubtlextSubletToString, 0);

  /* Class: tag */
  klass = rb_define_class_under(mod, "Tag", rb_cObject);
  rb_define_attr(klass,   "id",   1, 0);
  rb_define_attr(klass,   "name", 1, 0);

  rb_define_method(klass, "initialize", SubtlextTagInit,     1);
  rb_define_method(klass, "taggings",   SubtlextTagTaggings, 0);
  rb_define_method(klass, "save",       SubtlextTagSave,     0);
  rb_define_method(klass, "to_str",     SubtlextTagToString, 0);
  rb_define_alias(klass, "to_s", "to_str");

  /* Class: view */
  klass = rb_define_class_under(mod, "View", rb_cObject);
  rb_define_attr(klass,   "id",   1, 0);
  rb_define_attr(klass,   "win",  1, 0);
  rb_define_attr(klass,   "name", 1, 0);

  rb_define_method(klass, "initialize", SubtlextViewInit,          1);
  rb_define_method(klass, "clients",    SubtlextViewClientList,    0);
  rb_define_method(klass, "tags",       SubtlextViewTagList,       0);
  rb_define_method(klass, "tag",        SubtlextViewTagAdd,        1);
  rb_define_method(klass, "untag",      SubtlextViewTagDel,        1);
  rb_define_method(klass, "jump",       SubtlextViewJump,          0);
  rb_define_method(klass, "current?",   SubtlextViewCurrent,       0);
  rb_define_method(klass, "save",       SubtlextViewSave,          0);
  rb_define_method(klass, "to_str",     SubtlextViewToString,      0);
  rb_define_method(klass, "+",          SubtlextViewOperatorPlus,  1);
  rb_define_method(klass, "-",          SubtlextViewOperatorMinus, 1);
  rb_define_alias(klass, "to_s", "to_str");
} /* }}} */

// vim:ts=2:bs=2:sw=2:et:fdm=marker
